/*
 * http_get.c: Handles things associated with GET
 * 
 * Rob McCool
 * 
 */



#include "httpd.h"


int header_only;
static int num_includes;
static int allow;
static char allow_options;

void send_fd(FILE *f, FILE *fd, char *args);

/*
 * Thanks to Charles Henrich
 */
void process_include(FILE *f, FILE *fd, char *incstring, char *args)
{
    FILE *ifp;
    char srcfile[HUGE_STRING_LEN];
    char command[HUGE_STRING_LEN];
    char errstr[MAX_STRING_LEN];
    char cmd[10];

    if(sscanf(incstring,"%s \"%[^\"]", cmd, srcfile) != 2) {
        sprintf(errstr,"the include string %s was invalid",incstring);
        die(INCLUDE_ERROR,errstr,fd);
    }                

    if(strncasecmp(cmd,"srv",3) != 0) {
        fprintf(fd, "%s\r\n",cmd);
        return;
    }

    if(srcfile[0] != '|') {
        if(num_includes > MAXINCLUDES) {
            die(INCLUDE_ERROR,
                "the maximum number of includes has been exceeded",fd);
        }

        num_includes++;

        translate_name(srcfile,fd);
        ifp=fopen(srcfile,"r");

        if(ifp != NULL) {
            send_fd(ifp, fd, "");
            fclose(ifp);
        }
        else {
            sprintf(errstr,"the include file %s was invalid",srcfile);
            die(INCLUDE_ERROR,errstr,fd);
        }
    }
    else {
        if(strncasecmp(cmd,"srvurl",6) == 0) {
            sprintf(command,"%s '%s'",&srcfile[1],args);
        }
        else {
            strcpy(command,&srcfile[1]);
        }

        ifp=popen(command,"r");

        if(ifp != NULL) {
            send_fd(ifp, fd, args);
            pclose(ifp);
        }
        else {
            sprintf(errstr,"the command %s is invalid",&srcfile[1]);
            die(INCLUDE_ERROR,errstr,fd);
        }
    }
}

void send_fd_timed_out() {
    char errstr[MAX_STRING_LEN];

    sprintf(errstr,"httpd: send timed out for %s",remote_host);
    log_error(errstr);
    fclose(stdin);
    fclose(stdout);
    exit(0);
}

void send_fd(FILE *f, FILE *fd, char *args)
{
    int num_chars=0;
    char c;
    struct stat finfo;

    signal(SIGALRM,send_fd_timed_out);
    signal(SIGPIPE,send_fd_timed_out);
    alarm(timeout);

    if((allow_options & OPT_INCLUDES) && is_content_type("text/html")) {
        char *find="<inc";
        char incstring[MAX_STRING_LEN];
        int x,p;

        p=0;
        while(1) {
            c = fgetc(f);
            if(feof(f))
                return;
            if(tolower(c) == find[p]) {
                if((++p) == 4) {
                    x=0;
                    c=fgetc(f); /* get the space instead of the c */
                    while(c != '>') {
                        incstring[x++] = c;
                        c = fgetc(f);
                        if(feof(f)) {
                            incstring[x] = '\0';
                            fputs("<inc",fd);
                            fputs(incstring,fd);
                            return;
                        }
                    }
                    incstring[x] = '\0';
                    process_include(f,fd,incstring,args);
                    p=0;
                }
            }
            else {
                if(p) {
                    for(x=0;x<p;x++)
                        fputc(find[x],fd);
                    p=0;
                }
                fputc(c,fd);
            }
        }
    } else {
        while (1) {
            c = fgetc(f);
            if(feof(f))
                return;
            fputc(c,fd);
        }
    }
}

void send_file(char *file, FILE *fd, char *args) {
    FILE *f;
    struct stat finfo;

    if(!(f=fopen(file,"r"))) {
        unmunge_name(file);
        die(FORBIDDEN,file,fd); /* we've already established that it exists */
    }
    set_content_type(file);
    if(!assbackwards) {
        fstat(fileno(f),&finfo);
        set_content_length(finfo.st_size);
        set_last_modified(finfo.st_mtime);
        send_mime_headers(fd);
    }
    num_includes=0;
    if(!header_only) 
        send_fd(f,fd,args);
}

/* For scripts */
void set_env_vars() {
    char *t1,*t2,*t3;

    t1 = (char *)malloc(strlen("DOCUMENT_ROOT=") + strlen(document_root) + 1);
    sprintf(t1,"DOCUMENT_ROOT=%s",document_root);
    putenv(t1);
    t2 = (char *)malloc(strlen("SERVER_ROOT=") + strlen(server_root) + 1);
    sprintf(t2,"SERVER_ROOT=%s",server_root);
    putenv(t2);
    t3 = (char *)malloc(strlen("SERVER_NAME=") + strlen(server_hostname) + 6);
    sprintf(t3,"SERVER_NAME=%s:%d",server_hostname,port);
    putenv(t3);
}

void exec_get_script(char *path, char *args, FILE *fd) {
    FILE *tfp;
    struct stat finfo;
    int pid,pfd[2];
    char path_args[MAX_STRING_LEN];
    char t[MAX_STRING_LEN];
    register int n,x;

    path_args[0] = '\0';
    /* check if it's really a script with extra args */
    n=count_dirs(path);
    for(x=0;x<=n;x++) {
        make_dirstr(path,x+1,t);
        if(!(stat(t,&finfo))) {
            if(S_ISREG(finfo.st_mode)) {
                strcpy(path_args,&path[strlen(t)]);
                strcpy(path,t);
                goto run_script;
            }
        }
    }
    unmunge_name(path);
    die(NOT_FOUND,path,fd);
  run_script:
    if(!can_exec(&finfo)) {
        unmunge_name(path);
        die(FORBIDDEN,path,fd);
    }
    evaluate_access(path,&finfo,M_GET,&allow,&allow_options);
    if(!allow) {
        unmunge_name(path);
        die(FORBIDDEN,path,fd);
    }

    sprintf(t,"REMOTE_HOST=%s",remote_name);
    putenv(t);
    if(pipe(pfd) < 0)
        die(SERVER_ERROR,"could not open pipe",fd);

    signal(SIGALRM,send_fd_timed_out);
    signal(SIGPIPE,send_fd_timed_out);
    alarm(timeout);

    if((pid = fork()) < 0)
        die(SERVER_ERROR,"could not fork",fd);
    else if(!pid) {
        char *argv0;

        close(pfd[0]);
        if(pfd[1] != STDOUT_FILENO) {
            dup2(pfd[1],STDOUT_FILENO);
            close(pfd[1]);
        }
        if(argv0 = strrchr(path,'/'))
            argv0++;
        else
            argv0 = path;
        if(args[0] && path_args[0])
            execl(path,argv0,path_args,args,(char *)0);
        else if(args[0])
            execl(path,argv0,args,(char *)0);
        else if(path_args[0])
            execl(path,argv0,path_args,(char *)0);
        else
            execl(path,argv0,(char *)0);
    }
    else
        close(pfd[1]);

    tfp = fdopen(pfd[0],"r");

    scan_content_type(tfp,fd);
    if(location[0] == '/') {
        char *t = strdup(location);
        location[0] = '\0';
        send_node(t,NULL,fd);
        exit(0);
    }

    if(!assbackwards)
        send_mime_headers(fd);

    if(!header_only)
        send_fd(tfp,fd,args);
    fclose(tfp);
    waitpid(pid,NULL,0);
}

/* GET */
void send_node(char *file, char *args, FILE *fd)
{
    struct stat finfo;

    if(stat(file,&finfo) == -1) {
        unmunge_name(file);
        if(errno == ENOENT) die(NOT_FOUND,file,fd);
        die(FORBIDDEN,file,fd);
    }

    evaluate_access(file,&finfo,M_GET,&allow,&allow_options);

    if(!allow) {
        unmunge_name(file);
        die(FORBIDDEN,file,fd);
    }

    if(S_ISDIR(finfo.st_mode)) {
        char ifile[MAX_STRING_LEN];

        make_full_path(file,index_name,ifile);
        if(stat(ifile,&finfo) == -1) {
            if(allow_options & OPT_INDEXES)
                index_directory(file,fd);
            else {
                unmunge_name(file);
                die(FORBIDDEN,file,fd);
            }
        }
        else {
            if(file[strlen(file) - 1] != '/') {
                char url[MAX_STRING_LEN];
                strcpy_dir(ifile,file);
                unmunge_name(ifile);
                construct_url(url,ifile);
                die(REDIRECT,url,fd);
            }
            send_file(ifile,fd,args);
        }
        return;
    }
    if(S_ISREG(finfo.st_mode))
        send_file(file,fd,args);
    else {
        unmunge_name(file);
        die(FORBIDDEN,file,fd); /* device driver or pipe, no permission */
    }
}

