/*
 * http_put.c: Handles PUT and POST
 * 
 * Rob McCool
 * 
 */

#include "httpd.h"

static int allow;
static char allow_options;


void exec_post_script(char *path, char *args, FILE *in, FILE *out) {
    int pid, inpipe[2],outpipe[2], x;
    char cl[MAX_STRING_LEN];
    FILE *psin,*psout;

    sprintf(cl,"%d",content_length);

    if(pipe(inpipe) < 0)
        die(SERVER_ERROR,"httpd: could not create IPC pipe",out);
    if(pipe(outpipe) < 0)
        die(SERVER_ERROR,"httpd: could not create IPC pipe",out);
    if((pid = fork()) < 0)
        die(SERVER_ERROR,"httpd: could not fork new process",out);

    if(!pid) {
        char *argv0;

        if(outpipe[1] != STDOUT_FILENO) {
            dup2(outpipe[1],STDOUT_FILENO);
            close(outpipe[1]);
        }
        if(inpipe[0] != STDIN_FILENO) {
            dup2(inpipe[0],STDIN_FILENO);
            close(inpipe[0]);
        }
        if((argv0 = strrchr(path,'/')) != NULL)
            argv0++;
        else argv0 = path;
        execlp(path,argv0,cl,args,(char *)0);
    }
    else {
        close(outpipe[1]);
        close(inpipe[0]);
    }
    /* IS THIS NECESSARY??? */
    if(!(psout = fdopen(inpipe[1],"w")))
        die(SERVER_ERROR,"could not write to pipe",out);
    for(x=0;x<content_length;x++)
        fputc((char)fgetc(in),psout);
    fflush(psout);

    if(!(psin = fdopen(outpipe[0],"r")))
        die(SERVER_ERROR,"could not read from script",out);
    scan_content_type(psin,out);

    if(!assbackwards) {
/*        fstat(f,&finfo); */
/*        set_content_length(finfo.st_size); */
/*        set_last_modified(finfo.st_mtime); */
        send_mime_headers(out);
    }

    send_fd(psin,out,args);
    fclose(psin);
    fclose(psout);
    waitpid(pid,NULL,0);
}


void get_node(char *name, char *args, FILE *in, FILE *out, int script) {
    struct stat finfo;

    if(!script)
        die(NOT_IMPLEMENTED,"non-script",out); /* for now, script-only */

    if(strcmp(content_type,"application/x-www-form-urlencoded"))
        die(NOT_IMPLEMENTED,content_type,out); /* forms only */

    if(stat(name,&finfo) == -1) {
        unmunge_name(name);
        if(errno == ENOENT) die(NOT_FOUND,name,out);
        die(FORBIDDEN,name,out);
    }

    evaluate_access(name,&finfo,M_POST,&allow,&allow_options);

    if(!allow)
        die(FORBIDDEN,name,out);

    exec_post_script(name,args,in,out);
}
