/*
 * httpd.h: header for simple (ha! not anymore) http daemon
 */


/* Define one of these according to your system. */
#if defined(LINUX)
#undef BSD
#undef NO_KILLPG
#undef NO_SETSID

#elif defined(SUNOS4)
#define BSD
#undef NO_KILLPG
#undef NO_SETSID
char *crypt(char *pw, char *salt);

#elif defined(SOLARIS2)
#undef BSD
#undef NO_KILLPG
#undef NO_SETSID
#define bzero(a,b) memset(a,0,b)

#elif defined(IRIX)
#undef BSD
#undef NO_KILLPG
#undef NO_SETSID

#elif defined(HPUX)
#undef BSD
#define NO_KILLPG
#undef NO_SETSID

#elif defined(AIX)
#undef BSD
#undef NO_KILLPG
#undef NO_SETSID

#elif defined(ULTRIX)
#define BSD
#undef NO_KILLPG
#undef NO_SETSID
#define ULTRIX_BRAIN_DEATH
#define NEED_STRDUP

#elif defined(OSF1)
#define BSD
#undef NO_KILLPG
#undef NO_SETSID

#elif defined(SEQUENT)
#define BSD
#undef NO_KILLPG
#define NO_SETSID
#define NEED_STRDUP
#define tolower(c) (isupper(c) ? tolower(c) : c)

#elif defined(NEXT)
#define BSD
#undef NO_KILLPG
#define NO_SETSID
#define NEED_STRDUP
#define _POSIX_SOURCE
#define S_ISDIR(m)      (((m)&(S_IFMT)) == (S_IFDIR))
#define S_ISREG(m)      (((m)&(S_IFMT)) == (S_IFREG))

/* Unknown system - Edit these to match */
#else
/* BSD is whether your system uses BSD calls or System V calls. */
#define BSD
/* NO_KILLPG is set on systems that don't have killpg */
#undef NO_KILLPG
/* NO_SETSID is set on systems that don't have setsid */
#undef NO_SETSID
/* NEED_STRDUP is set on stupid systems that don't have strdup. */
#undef NEED_STRDUP
#endif

/*
 * The particular directory style your system supports. If you have dirent.h
 * in /usr/include (POSIX) or /usr/include/sys (SYSV), #include 
 * that file and define DIR_TYPE to be dirent. Otherwise, if you have 
 * /usr/include/sys/dir.h, define DIR_TYPE to be direct and include that
 * file. If you have neither, I'm confused.
 */
#ifndef NEXT
#include <dirent.h>
#define DIR_TYPE dirent
#else
#include <sys/dir.h>
#define DIR_TYPE direct
#endif

/* ----------------------------- config dir ------------------------------ */

/* Define this to be the default server home dir. Anything later in this
 * file with a relative pathname will have this added.
 */
#define HTTPD_ROOT "/usr/local/etc/httpd"

/* Root of server */
#define DOCUMENT_LOCATION "/usr/local/etc/httpd/htdocs"

/* Max. number of aliases */
#define MAX_ALIASES 20

/* Max. number of security defines */
#define MAX_SECURITY 50

/* Max. number of include files */
#define MAXINCLUDES 20

/* Default administrator's address */
#define DEFAULT_ADMIN "[no address given]"

/* 
 * --------- You shouldn't have to edit anything below this line ----------
 *
 * Any modifications to any defaults not defined above should be done in the 
 * respective config. file. 
 *
 */


/* -------------- Port number for server running standalone --------------- */

#define DEFAULT_PORT 80

/* --------- Default user name and group name running standalone ---------- */
/* --- These may be specified as numbers by placing a # before a number --- */

#define DEFAULT_USER "nobody"
#define DEFAULT_GROUP "nobody"

/* The name of the log files */
#define DEFAULT_XFERLOG "logs/access_log"
#define DEFAULT_ERRORLOG "logs/error_log"
#define DEFAULT_PIDLOG "logs/httpd.pid"

/* Define this to be what your HTML directory content files are called */
#define DEFAULT_INDEX "index.html"

/* Define this to be what type you'd like returned for files with unknown */
/* suffixes */
#define DEFAULT_TYPE "text/html"

/* Define this to be what your per-directory security files are called */
#define DEFAULT_ACCESS_FNAME ".htaccess"

/* The name of the server config file */
#define SERVER_CONFIG_FILE "conf/httpd.conf"

/* The name of the document config file */
#define RESOURCE_CONFIG_FILE "conf/srm.conf"

/* The name of the MIME types file */
#define TYPES_CONFIG_FILE "conf/mime.types"

/* The name of the access file */
#define ACCESS_CONFIG_FILE "conf/access.conf"


/* The default directory in user's home dir */
#define DEFAULT_USER_DIR "public_html"


/* The default string lengths */
#define MAX_STRING_LEN 256
#define HUGE_STRING_LEN 8192

/* The timeout for waiting for messages */
#define DEFAULT_TIMEOUT 1200


/* ------------------------------ error types ------------------------------ */

#define SERVER_VERSION "NCSA/1.0a5"
#define SERVER_PROTOCOL "HTTP/1.0"
#define SERVER_SUPPORT "httpd@ncsa.uiuc.edu"

#define REDIRECT 302
#define BAD_REQUEST 400
#define AUTH_REQUIRED 401
#define FORBIDDEN 403
#define NOT_FOUND 404
#define SERVER_ERROR 500
#define NOT_IMPLEMENTED 501
#define INCLUDE_ERROR 6991

#define METHODS 3
#define M_GET 0
#define M_PUT 1
#define M_POST 2

#define OPT_NONE 0
#define OPT_INDEXES 1
#define OPT_INCLUDES 2
#define OPT_SYM_LINKS 4
#define OPT_ALL (OPT_INDEXES|OPT_INCLUDES|OPT_SYM_LINKS)

#define OR_NONE 0
#define OR_LIMIT 1
#define OR_OPTIONS 2
#define OR_FILEINFO 4
#define OR_OVERRIDE 8
#define OR_AUTHCFG 16
#define OR_ALL (OR_LIMIT|OR_OPTIONS|OR_FILEINFO|OR_OVERRIDE|OR_AUTHCFG)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <ctype.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>  /* for inet_ntoa */
#include <time.h>  /* for ctime */
#include <signal.h>
#include <errno.h>
#include <sys/wait.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>

#ifdef ultrix
#define ULTRIX_BRAIN_DEATH
#endif

/* Just in case your linefeed isn't the one the other end is expecting. */
#define LF 10
#define CR 13

/* For access control */
#define DENY_THEN_ALLOW 0
#define ALLOW_THEN_DENY 1

/* Struct shared by access and auth */
typedef struct {
    char *d;
    char opts;
    char override;

    int order[METHODS];

    int num_allow[METHODS];
    char *allow[METHODS][MAX_SECURITY];
    int num_auth[METHODS];
    char *auth[METHODS][MAX_SECURITY];

    char *auth_type;
    char *auth_name;
    char *auth_pwfile;
    char *auth_grpfile;

    int num_deny[METHODS];
    char *deny[METHODS][MAX_SECURITY];
} security_data;

/* Globals */

/* Server config */
extern int standalone;
extern int port;
extern uid_t user_id;
extern gid_t group_id;
extern char server_root[MAX_STRING_LEN];
extern char error_fname[MAX_STRING_LEN];
extern char xfer_fname[MAX_STRING_LEN];
extern char pid_fname[MAX_STRING_LEN];
extern char server_admin[MAX_STRING_LEN];
extern char *server_hostname;
extern char srm_confname[MAX_STRING_LEN];
extern char access_confname[MAX_STRING_LEN];
extern char types_confname[MAX_STRING_LEN];
extern int timeout;

/* Document config */
extern char user_dir[MAX_STRING_LEN];
extern char index_name[MAX_STRING_LEN];
extern char access_name[MAX_STRING_LEN];
extern char document_root[MAX_STRING_LEN];
extern char default_type[MAX_STRING_LEN];

/* Security config */
extern int num_sec;
extern security_data sec[MAX_SECURITY];

/* Auth config */
extern char *auth_type;
extern char *auth_name;
extern char *auth_pwfile;
extern char *auth_grpfile;

/* Request information */
extern char *remote_name;
extern int assbackwards;
extern int header_only;
extern char *remote_host;
extern char *remote_ip;
extern char *remote_name;

/* MIME */
extern char auth_line[MAX_STRING_LEN];
extern int content_length;
extern char content_type[MAX_STRING_LEN];
extern char location[MAX_STRING_LEN];

/* Function prototypes. */

/* http_config */
extern void read_config();
extern void parse_htaccess(char *file, char override);
extern int get_pw(char *user, char *pw);
extern int in_group(char *user, char *group);
extern int init_group(char *grpfile);
extern void kill_group();

/* http_alias */
extern void reset_aliases();
extern void dump_aliases();
extern void add_alias(char *f, char *r, int is_script);
extern int translate_name(char *name,FILE *fd);
extern void unmunge_name(char *name);

/* http_request */
extern void process_request(FILE *in, FILE *out);

/* http_get */
extern void send_file(char *file,FILE *fd, char *args);
extern void process_include(FILE *f, FILE *fd, char *incstring, char *args);
extern void send_node(char *name, char *args, FILE *fd);
extern void exec_get_script(char *path, char *args, FILE *fd);
extern void set_env_vars();

/* http_put */
extern void get_node(char *name, char *args, FILE *in, FILE *out, int script);

/* http_dir */
extern void index_directory(char *name, FILE *fd);

/* http_log */
extern void log_transaction(char *cmd_line);
extern void log_error(char *err);
extern void die(int type, char *err_string, FILE *fd);
extern void open_logs();
extern void close_logs();

/* http_mime */
extern void get_mime_headers(FILE *fd);
extern void send_mime_headers(FILE *fd);
extern void set_content_type(char *fn);
extern int scan_content_type(FILE *f, FILE *fd);
extern void add_content_type(char *fn, char *t);
extern void set_content_length(int l);
extern void dump_types();
extern void init_mime();
extern void kill_mime();
extern int is_content_type(char *type);
extern void dump_default_header(FILE *fd);

/* http_access */
extern void evaluate_access(char *path, struct stat *finfo,int m, int *allow, 
                            char *op);
extern void kill_security();

/* http_auth */
extern void check_auth(security_data *s, int m);

/* util */
extern void strsubfirst(int start,char *dest, char *src);
extern void make_full_path(char *src1,char *src2,char *dst);
extern int is_directory(char *name);
extern void getparents(char *name);
extern uid_t uname2id(char *name);
extern gid_t gname2id(char *name);
extern int getline(char *s, int n, FILE *f, unsigned int timeout);
extern void getword(char *word, char *line, char stop);
extern void get_remote_host(int fd);
extern char *get_time();
extern char *get_gmt();
extern void make_dirstr(char *s, int n, char *d);
extern int count_dirs(char *path);
extern void strcpy_dir(char *d, char *s);
extern void unescape_url(char *url);
extern void escape_url(char *url);
extern void escape_shell_cmd(char *cmd);
extern void plustospace(char *str);
extern void str_tolower(char *str);
extern void uudecode(char *s,unsigned char *d,int dl);
#ifdef NEED_STRDUP
extern char *strdup (char *str);
#endif
int ind(char *s, char c);
int rind(char *s, char c);
void construct_url(char *d, char *s);
void get_local_host();
int get_portnum(int sd,FILE *out);
int can_exec(struct stat *finfo);
